@@ -12,12 +12,18 @@
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
-# ==============================================================================
+# ============================================================================
+"""Module of visualization tools.
 
-"""Module for visualizing embeddings.
+Functions:
+    bvn_ellipse: Create an ellipse representing a bivariate normal
+        distribution.
+    heatmap_embeddings: Create a heatmap of embeddings.
+    embedding_output_dimension: Visualize embedding values for a
+        requested output dimension.
+    embedding_input_dimension: Visualize embedding values for a
+        requested input dimension.
 
-Todo:
-    classes -> class_dict
 """
 
 import os
@@ -27,6 +33,226 @@ import matplotlib
 import matplotlib.pyplot as plt
 import numpy as np
 from PIL import Image
+import tensorflow_probability as tfp
+
+
+def heatmap_embeddings(fig, ax, embedding, group_idx=0, cmap=None):
+    """Visualize embeddings as a heatmap.
+
+    Intended to handle rank 2 and rank 3 embeddings.
+
+    Arguments:
+        fig: A Matplotlib Figure object.
+        ax: A Matplotlib Axes object.
+        embedding: An embedding layer.
+        cmap (optional): A Matplotlib compatible colormap.
+
+    """
+    if cmap is None:
+        cmap = matplotlib.cm.get_cmap('Greys')
+
+    if isinstance(embedding.embeddings, tfp.distributions.Distribution):
+        # Handle distribution.
+        z_mode = embedding.embeddings.mode().numpy()
+    else:
+        # Handle point estimate.
+        z_mode = embedding.embeddings.numpy()
+
+    rank = z_mode.ndim
+    if rank == 3:
+        z_mode = z_mode[group_idx]
+    if embedding.mask_zero:
+        z_mode = z_mode[1:]
+
+    n_dim = z_mode.shape[-1]
+    z_mode_max = np.max(z_mode)
+    im = ax.imshow(
+        z_mode, cmap=cmap, interpolation='none', vmin=0., vmax=z_mode_max
+    )
+    # TODO use matshow to fix interpolation issue.
+
+    # Note: imshow displays different rows as different values of y and
+    # different columns as different values of x.
+    ax.set_xticks([0, n_dim-1])
+    ax.set_xticklabels([0, n_dim-1])
+    ax.set_xlabel('Dimension')
+    ax.set_ylabel('Stimulus')
+    fig.colorbar(im, ax=ax)
+
+
+def embedding_output_dimension(fig, ax, embedding, idx, group_idx=0, c='b'):
+    """Visualize embedding values for a requested output dimension.
+
+    Plots point estimates of embedding values for the requested
+    output dimension.
+
+    If the embedding layer is a distribution, also attempts to draw a
+    thick linewidth interval indicating the middle 50% probability mass
+    and a thin linewidth interval indicating the middle 99% probability
+    mass via the inverse CDF function.
+
+    Intended to handle rank 2 and rank 3 embeddings.
+
+    Arguments:
+        fig: A Matplotlib Figure object.
+        ax: A Matplotlib Axes object.
+        embedding: An embedding layer.
+        idx: Index of requested output dimension to visualize.
+        c (optional): Color of interval marks.
+
+    """
+    if isinstance(embedding.embeddings, tfp.distributions.Distribution):
+        z_mode = embedding.embeddings.mode().numpy()
+    else:
+        z_mode = embedding.embeddings.numpy()
+
+    rank = z_mode.ndim
+    if rank == 3:
+        z_mode = z_mode[group_idx]
+
+    y_max = np.max(z_mode)
+    z_mode = z_mode[:, idx]
+
+    # Handle masking.
+    if embedding.mask_zero:
+        z_mode = z_mode[1:]
+
+    n_input_dim = z_mode.shape[0]
+
+    # Scatter point estimate.
+    xg = np.arange(n_input_dim)
+    ax.scatter(xg, z_mode, c=c, marker='_', linewidth=1)
+
+    if hasattr(embedding, 'posterior'):
+        dist = embedding.posterior.embeddings.distribution
+
+        # Middle 99% probability mass.
+        p = .99
+        v = (1 - p) / 2
+        quant_lower = dist.quantile(v).numpy()[:, idx]
+        quant_upper = dist.quantile(1-v).numpy()
+        y_max = np.max(quant_upper)
+        quant_upper = quant_upper[:, idx]
+
+        # Middle 50% probability mass.
+        p = .5
+        v = (1 - p) / 2
+        mid_lower = dist.quantile(v).numpy()[:, idx]
+        mid_upper = dist.quantile(1-v).numpy()[:, idx]
+
+        if embedding.posterior.mask_zero:
+            quant_lower = quant_lower[1:]
+            quant_upper = quant_upper[1:]
+            mid_lower = mid_lower[1:]
+            mid_upper = mid_upper[1:]
+
+        for i_dim in range(n_input_dim):
+            xg = np.array([i_dim, i_dim])
+            yg = np.array(
+                [quant_lower[i_dim], quant_upper[i_dim]]
+            )
+            ax.plot(xg, yg, c=c, linewidth=1)
+
+            yg = np.array(
+                [mid_lower[i_dim], mid_upper[i_dim]]
+            )
+            ax.plot(xg, yg, c=c, linewidth=3)
+
+    ax.set_xlabel('Input Dimension')
+    ax.set_xlim([-.5, n_input_dim-.5])
+
+    ax.set_ylabel(r'$x$')
+    ax.set_ylim([0, 1.05 * y_max])
+    ax.set_yticks([0, 1.05 * y_max])
+    ax.set_yticklabels(['0', '{0:.1f}'.format(1.05 * y_max)])
+
+
+def embedding_input_dimension(fig, ax, embedding, idx, group_idx=0, c='b'):
+    """Visualize embedding values for a requested input dimension.
+
+    Plots point estimates of embedding values for the requested
+    input dimension.
+
+    Intended to handle rank 2 and rank 3 embeddings.
+
+    If the embedding layer is a distribution, also attempts to draw a
+    thick linewidth interval indicating the middle 50% probability mass
+    and a thin linewidth interval indicating the middle 99% probability
+    mass via the inverse CDF function.
+
+    Arguments:
+        fig: A Matplotlib Figure object.
+        ax: A Matplotlib Axes object.
+        embedding: An embedding layer.
+        idx: Index of requested input dimension to visualize.
+        c (optional): Color of interval marks.
+
+    """
+    if isinstance(embedding.embeddings, tfp.distributions.Distribution):
+        z_mode = embedding.embeddings.mode().numpy()
+    else:
+        z_mode = embedding.embeddings.numpy()
+
+    rank = z_mode.ndim
+    if rank == 3:
+        z_mode = z_mode[group_idx]
+
+    y_max = np.max(z_mode)
+    z_mode = z_mode[idx, :]
+
+    # Handle masking.
+    if embedding.mask_zero:
+        z_mode = z_mode[1:]
+
+    n_output_dim = z_mode.shape[0]
+
+    # Scatter point estimate.
+    xg = np.arange(n_output_dim)
+    ax.scatter(xg, z_mode, c=c, marker='_')
+
+    # Add posterior quantiles if available.
+    if hasattr(embedding, 'posterior'):
+        dist = embedding.posterior.embeddings.distribution
+
+        # Middle 99% of probability mass.
+        p = .99
+        v = (1 - p) / 2
+        quant_lower = dist.quantile(v).numpy()[idx, :]
+        quant_upper = dist.quantile(1-v).numpy()
+        y_max = np.max(quant_upper)
+        quant_upper = quant_upper[idx, :]
+
+        # Middle 50% probability mass.
+        p = .5
+        v = (1 - p) / 2
+        mid_lower = dist.quantile(v).numpy()[idx, :]
+        mid_upper = dist.quantile(1-v).numpy()[idx, :]
+
+        if embedding.posterior.mask_zero:
+            quant_lower = quant_lower[1:]
+            quant_upper = quant_upper[1:]
+            mid_lower = mid_lower[1:]
+            mid_upper = mid_upper[1:]
+
+        for i_dim in range(n_output_dim):
+            xg = np.array([i_dim, i_dim])
+            yg = np.array(
+                [quant_lower[i_dim], quant_upper[i_dim]]
+            )
+            ax.plot(xg, yg, c=c, linewidth=1)
+
+            yg = np.array(
+                [mid_lower[i_dim], mid_upper[i_dim]]
+            )
+            ax.plot(xg, yg, c=c, linewidth=3)
+
+    ax.set_xlabel('Output Dimension')
+    ax.set_xlim([-.5, n_output_dim-.5])
+
+    ax.set_ylabel(r'$x$')
+    ax.set_ylim([0, 1.05 * y_max])
+    ax.set_yticks([0, 1.05 * y_max])
+    ax.set_yticklabels(['0', '{0:.1f}'.format(1.05 * y_max)])
 
 
 def visualize_embedding_static(
@@ -407,3 +633,38 @@ def image_2d(
         # in plot.
         for i_handle in lgnd.legendHandles:
             i_handle._sizes = [30]
+
+
+def bvn_ellipse(loc, cov, r=1.96, **kwargs):
+    """Return ellipse object representing bivariate normal.
+
+    This code was inspired by a solution posted on Stack Overflow:
+    https://stackoverflow.com/a/25022642/1860294
+
+    Arguments:
+        loc: A 1D array denoting the mean.
+        cov: A 2D array denoting the covariance matrix.
+        r (optional): The radius (specified in standard deviations) at
+            which to draw the ellipse. The default value corresponds to
+            an ellipse indicating a region containing 95% of the
+            probability mass.
+        kwargs (optional): Additional key-word arguments to pass to
+            `matplotlib.patches.Ellipse` constructor.
+
+    Returns:
+        ellipse: A `matplotlib.patches.Ellipse` artist object.
+
+    """
+    def eig_sorted(cov):
+        """Sort eigenvalues."""
+        vals, vecs = np.linalg.eigh(cov)
+        order = vals.argsort()[::-1]
+        return vals[order], vecs[:, order]
+
+    vals, vecs = eig_sorted(cov)
+    theta = np.degrees(np.arctan2(*vecs[:, 0][::-1]))
+    w, h = 2 * r * np.sqrt(vals)
+    ellipse = matplotlib.patches.Ellipse(
+        xy=(loc[0], loc[1]), width=w, height=h, angle=theta, **kwargs
+    )
+    return ellipse
